# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Ticker(Component):
    """A Ticker component.
A light wrapper of ReactTicker.

Keyword arguments:

- children (a list of or a singular dash component, string or number; optional):
    The children of this component.

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- className (string; optional):
    The class of the component.

- direction (a value equal to: "toRight", "toLeft"; optional)

- height (number | string; optional)

- mode (a value equal to: "chain", "await", "smooth"; optional)

- move (boolean; optional)

- offset (number | string; optional)

- speed (number; optional)"""
    @_explicitize_args
    def __init__(self, children=None, direction=Component.UNDEFINED, mode=Component.UNDEFINED, move=Component.UNDEFINED, offset=Component.UNDEFINED, speed=Component.UNDEFINED, height=Component.UNDEFINED, id=Component.UNDEFINED, className=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'id', 'className', 'direction', 'height', 'mode', 'move', 'offset', 'speed']
        self._type = 'Ticker'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'id', 'className', 'direction', 'height', 'mode', 'move', 'offset', 'speed']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Ticker, self).__init__(children=children, **args)



import functools
import hashlib
import json
import pickle
import secrets
import uuid
import dash_core_components as dcc
import dash
import dash_html_components as html
import dash.dependencies as dd
import plotly

from dash.dependencies import Input, Output, State, MATCH, ALL, ALLSMALLER, _Wildcard, ClientsideFunction
from dash.development.base_component import Component
from flask import session
from flask_caching.backends import FileSystemCache, RedisCache
from more_itertools import flatten
from collections import defaultdict
from typing import Dict

_wildcard_mappings = {ALL: "<ALL>", MATCH: "<MATCH>", ALLSMALLER: "<ALLSMALLER>"}
_wildcard_values = list(_wildcard_mappings.values())


# region Dash proxy

class DashProxy(dash.Dash):

    def __init__(self, *args, transforms=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.callbacks = []
        self.clientside_callbacks = []
        self.arg_types = [dd.Output, dd.Input, dd.State]
        self.transforms = transforms if transforms is not None else []
        # Do the transform initialization.
        for transform in self.transforms:
            transform.init(self)

    def _collect_callback(self, *args, **kwargs):
        """
         This method saves the callbacks on the DashTransformer object. It acts as a proxy for the Dash app callback.
        """
        # Parse Output/Input/State (could be made simpler by enforcing input structure)
        keys = ['output', 'inputs', 'state']
        args = list(args) + list(flatten([_extract_list_from_kwargs(kwargs, key) for key in keys]))
        callback = {arg_type: [] for arg_type in self.arg_types}
        arg_order = []
        multi_output = False
        for arg in args:
            elements = _as_list(arg)
            for element in elements:
                for key in callback:
                    if isinstance(element, key):
                        # Check if this is a wild card output.
                        if not multi_output and isinstance(element, dd.Output):
                            component_id = element.component_id
                            if isinstance(component_id, dict):
                                multi_output = any([component_id[k] in [dd.ALLSMALLER, dd.ALL] for k in component_id])
                        callback[key].append(element)
                        arg_order.append(element)
        if not multi_output:
            multi_output = len(callback[dd.Output]) > 1
        # Save the kwargs for later.
        callback["kwargs"] = kwargs
        callback["sorted_args"] = arg_order
        callback["multi_output"] = multi_output

        return callback

    def callback(self, *args, **kwargs):
        """
         This method saves the callbacks on the DashTransformer object. It acts as a proxy for the Dash app callback.
        """
        callback = self._collect_callback(*args, **kwargs)
        self.callbacks.append(callback)

        def wrapper(f):
            callback["f"] = f

        return wrapper

    def clientside_callback(self, clientside_function, *args, **kwargs):
        callback = self._collect_callback(*args, **kwargs)
        callback["f"] = clientside_function
        self.clientside_callbacks.append(callback)

    def _register_callbacks(self, app=None):
        callbacks, clientside_callbacks = self._resolve_callbacks()
        app = super() if app is None else app
        for cb in callbacks:
            outputs = cb[dd.Output][0] if len(cb[dd.Output]) == 1 else cb[dd.Output]
            app.callback(outputs, cb[dd.Input], cb[dd.State], **cb["kwargs"])(cb["f"])
        for cb in clientside_callbacks:
            outputs = cb[dd.Output][0] if len(cb[dd.Output]) == 1 else cb[dd.Output]
            app.clientside_callback(cb["f"], outputs, cb[dd.Input], cb[dd.State], **cb["kwargs"])

    def _layout_value(self):
        layout = self._layout() if self._layout_is_function else self._layout
        for transform in self.transforms:
            layout = transform.layout(layout, self._layout_is_function)
        return layout

    def _setup_server(self):
        """
         This method registers the callbacks on the Dash app and injects a session secret.
        """
        # Register the callbacks.
        self._register_callbacks()
        # Proceed as normally.
        super()._setup_server()
        # Set session secret. Used by some subclasses.
        if not self.server.secret_key:
            self.server.secret_key = secrets.token_urlsafe(16)

    def _resolve_callbacks(self):
        """
         This method resolves the callbacks, i.e. it applies the callback injections.
        """
        callbacks, clientside_callbacks = self.callbacks, self.clientside_callbacks
        for transform in self.transforms:
            callbacks, clientside_callbacks = transform.apply(callbacks, clientside_callbacks)
        return callbacks, clientside_callbacks

    def hijack(self, app: dash.Dash):
        # Change properties.
        app.config.update(self.config)
        app.title = self.title
        app.index_string = self.index_string
        # Inject layout.
        app.layout = html.Div()  # fool layout validator
        app._layout_value = self._layout_value
        # Register callbacks.
        self._register_callbacks(app)
        # Setup secret.
        if not app.server.secret_key:
            app.server.secret_key = secrets.token_urlsafe(16)


def _get_session_id(session_key=None):
    session_key = "session_id" if session_key is None else session_key
    # Create unique session id.
    if not session.get(session_key):
        session[session_key] = secrets.token_urlsafe(16)
    return session.get(session_key)


def _as_list(item):
    if item is None:
        return []
    return item if isinstance(item, list) else [item]


def _create_callback_id(item):
    cid = item.component_id
    if isinstance(cid, dict):
        cid = {key: cid[key] if cid[key] not in _wildcard_mappings else _wildcard_mappings[cid[key]] for key in cid}
        cid = json.dumps(cid)
    return "{}.{}".format(cid, item.component_property)


def _extract_list_from_kwargs(kwargs: dict, key: str) -> list:
    if kwargs is not None and key in kwargs:
        contents = kwargs.pop(key)
        if contents is None:
            return []
        if isinstance(contents, list):
            return contents
        else:
            return [contents]
    else:
        return []


def plotly_jsonify(data):
    return json.loads(json.dumps(data, cls=plotly.utils.PlotlyJSONEncoder))


class DashTransform:

    def init(self, dt):
        pass

    def apply(self, callbacks, clientside_callbacks):
        return self.apply_serverside(callbacks), self.apply_clientside(clientside_callbacks)

    def apply_serverside(self, callbacks):
        return callbacks  # per default do nothing

    def apply_clientside(self, callbacks):
        return callbacks  # per default do nothing

    def layout(self, layout, layout_is_function):
        return layout


# endregion

# region Prefix ID transform

class PrefixIdTransform(DashTransform):

    def __init__(self, prefix, prefix_func=None):
        self.prefix = prefix
        self.prefix_func = prefix_func if prefix_func is not None else prefix_component
        self.initialized = False

    def _apply(self, callbacks):
        for callback in callbacks:
            for arg in callback["sorted_args"]:
                arg.component_id = apply_prefix(self.prefix, arg.component_id)
        return callbacks

    def apply_serverside(self, callbacks):
        return self._apply(callbacks)

    def apply_clientside(self, callbacks):
        return self._apply(callbacks)

    def layout(self, layout, layout_is_function):
        # TODO: Will this work with layout functions?
        if layout_is_function or not self.initialized:
            prefix_recursively(layout, self.prefix, self.prefix_func)
            self.initialized = True
        return layout


def apply_prefix(prefix, component_id):
    if isinstance(component_id, dict):
        for key in component_id:
            # This branch handles the IDs. TODO: Can we always assume use of ints?
            if type(component_id[key]) == int:
                continue
            # This branch handles the wildcard callbacks.
            if isinstance(component_id[key], _Wildcard):
                continue
            # All "normal" props are prefixed.
            component_id[key] = "{}-{}".format(prefix, component_id[key])
        return component_id
    return "{}-{}".format(prefix, component_id)


def prefix_recursively(item, key, prefix_func):
    prefix_func(key, item)
    if hasattr(item, "children"):
        children = _as_list(item.children)
        for child in children:
            prefix_recursively(child, key, prefix_func)


def prefix_component(key, component):
    if hasattr(component, "id"):
        component.id = apply_prefix(key, component.id)
    if not hasattr(component, "_namespace"):
        return
    # Special handling of dash bootstrap components. TODO: Maybe add others?
    if component._namespace == "dash_bootstrap_components":
        if component._type == "Tooltip":
            component.target = apply_prefix(key, component.target)


# endregion

# region Trigger transform (the only default transform)

class Trigger(Input):
    """
     Like an Input, a trigger can trigger a callback, but it's values it not included in the resulting function call.
    """

    def __init__(self, component_id, component_property):
        super().__init__(component_id, component_property)


class TriggerTransform(DashTransform):

    # NOTE: This transform cannot be implemented for clientside callbacks since the JS can't be modified from here.

    def apply_serverside(self, callbacks):
        for callback in callbacks:
            is_trigger = trigger_filter(callback["sorted_args"])
            # Check if any triggers are there.
            if not any(is_trigger):
                continue
            # If so, filter the callback args.
            f = callback["f"]
            callback["f"] = filter_args(is_trigger)(f)
        return callbacks


def filter_args(args_filter):
    def wrapper(f):
        @functools.wraps(f)
        def decorated_function(*args):
            filtered_args = [arg for j, arg in enumerate(args) if not args_filter[j]]
            return f(*filtered_args)

        return decorated_function

    return wrapper


def trigger_filter(args):
    inputs_args = [item for item in args if isinstance(item, dd.Input) or isinstance(item, dd.State)]
    is_trigger = [isinstance(item, Trigger) for item in inputs_args]
    return is_trigger


# endregion

# region Multiplexer transform

def _mp_id(output: Output, idx: int) -> Dict[str, str]:
    if isinstance(output.component_id, dict):
        return {**output.component_id, **dict(prop=output.component_property, idx=idx)}
    return dict(id=output.component_id, prop=output.component_property, idx=idx)


def _escape_wildcards(mp_id):
    if not isinstance(mp_id, dict):
        return mp_id
    for key in mp_id:
        # The ALL wildcard is supported.
        if mp_id[key] == ALL:
            mp_id[key] = _wildcard_mappings[ALL]
            continue
        # Other ALL wildcards are NOT supported.
        if mp_id[key] in _wildcard_mappings:
            raise ValueError(f"Multiplexer does not support wildcard [{mp_id[key]}]")
    return mp_id


def _mp_element(mp_id: Dict[str, str]) -> dcc.Store:
    return dcc.Store(id=mp_id)


def _mp_prop() -> str:
    return "data"


def inject_proxies_recursively(node, proxy_map):
    if not hasattr(node, "children") or node.children is None:
        return
    children = _as_list(node.children)
    modified = False
    for i, child in enumerate(children):
        # Do recursion.
        inject_proxies_recursively(child, proxy_map)
        # Attach the proxy components as children of the original component to ensure dcc.Loading works.
        if not hasattr(child, "id"):
            continue
        for key in proxy_map:
            if not child.id == key:
                continue
            children[i] = html.Div([child] + proxy_map[key])
            modified = True
    if modified:
        node.children = children


class MultiplexerTransform(DashTransform):
    """
    The MultiplexerTransform makes it possible to target an output by callbacks multiple times. Under the hood, proxy
    components (dcc.Store) are used, and the proxy_location keyword argument determines where these proxies are placed.
    The default value "inplace" means that the original component is replace by a Div element that wraps the original
    component and its proxies. The means that dcc.Loading will work, but it also means that if you replace the layout
    of a component higher in the tree in a callback, you might end up deleting the proxies (!), which will break your
    app. For this particular case, you can set proxy_location to a custom component (or None to use the layout root),
    to place the proxies here instead. To use dcc.Loading for this particular case, the proxy_location must be wrapped.
    """

    def __init__(self, proxy_location=None, proxy_wrapper_map=None):
        self.initialized = False
        self.proxy_location = proxy_location
        self.proxy_map = defaultdict(lambda: [])
        self.proxy_wrapper_map = proxy_wrapper_map
        self.app = DashProxy()

    def layout(self, layout, layout_is_function):
        if layout_is_function or not self.initialized:
            # Apply wrappers if needed.
            if self.proxy_wrapper_map:
                for key in self.proxy_wrapper_map:
                    if key in self.proxy_map:
                        self.proxy_map[key] = _as_list(self.proxy_wrapper_map[key](self.proxy_map[key]))
            # Inject proxies in a user defined component.
            if self.proxy_location == "inplace":
                inject_proxies_recursively(layout, self.proxy_map)
            # Inject proxies in a component, either user defined or top level.
            else:
                target = self.proxy_location if isinstance(self.proxy_location, Component) else layout
                proxies = list(flatten(list(self.proxy_map.values())))
                target.children = _as_list(target.children) + proxies
            self.initialized = True
        return layout

    def apply(self, callbacks, clientside_callbacks):
        all_callbacks = callbacks + clientside_callbacks
        # Group by output.
        output_map = defaultdict(list)
        for callback in all_callbacks:
            for output in callback[Output]:
                output_map[output].append(callback)
        # Apply multiplexer where needed.
        for output in output_map:
            # If there is only one output, multiplexing is not needed.
            if len(output_map[output]) == 1:
                continue
            self._apply_multiplexer(output, output_map[output])

        return callbacks, clientside_callbacks + self.app.clientside_callbacks

    def _apply_multiplexer(self, output, callbacks):
        inputs = []
        proxies = []
        for i, callback in enumerate(callbacks):
            mp_id = _mp_id(output, i)
            mp_id_escaped = _escape_wildcards(mp_id)
            # Create proxy element.
            proxies.append(_mp_element(mp_id_escaped))
            # Assign proxy element as output.
            callback[Output][callback[Output].index(output)] = Output(mp_id_escaped, _mp_prop())
            # Create proxy input.
            inputs.append(Input(mp_id, _mp_prop()))
        # Collect proxy elements to add to layout.
        self.proxy_map[output].extend(proxies)
        # Create multiplexer callback. Clientside for best performance. TODO: Is this robust?
        self.app.clientside_callback("""
            function(){
                const ts = dash_clientside.callback_context.triggered;
                return ts[0].value;
            }
        """, output, inputs, prevent_initial_call=True)


# endregion

# region Server side output transform

class EnrichedOutput(Output):
    """
     Like a normal Output, includes additional properties related to storing the data.
    """

    def __init__(self, component_id, component_property, backend=None, session_check=None, arg_check=True):
        super().__init__(component_id, component_property)
        self.backend = backend
        self.session_check = session_check
        self.arg_check = arg_check


class ServersideOutput(EnrichedOutput):
    """
     Like a normal Output, but with the content stored only server side.
    """


class ServersideOutputTransform(DashTransform):

    def __init__(self, backend=None, session_check=True, arg_check=True):
        self.backend = backend if backend is not None else FileSystemStore()
        self.session_check = session_check
        self.arg_check = arg_check

    def init(self, dt):
        # Set session secret (if not already set).
        if not dt.server.secret_key:
            dt.server.secret_key = secrets.token_urlsafe(16)

    # NOTE: Doesn't make sense for clientside callbacks.

    def apply_serverside(self, callbacks):
        # 1) Create index.
        serverside_callbacks = []
        serverside_output_map = {}
        for callback in callbacks:
            # If memoize keyword is used, serverside caching is needed.
            memoize = callback["kwargs"].get("memoize", None)
            serverside = False
            # Keep tract of which outputs are server side outputs.
            for output in callback[dd.Output]:
                if isinstance(output, ServersideOutput):
                    serverside_output_map[_create_callback_id(output)] = output
                    serverside = True
                # Set default values.
                if not isinstance(output, ServersideOutput) and not memoize:
                    continue
                output.backend = output.backend if output.backend is not None else self.backend
                output.arg_check = output.arg_check if output.arg_check is not None else self.arg_check
                output.session_check = output.session_check if output.session_check is not None else self.session_check
            # Keep track of server side callbacks.
            if serverside or memoize:
                serverside_callbacks.append(callback)
        # 2) Inject cached data into callbacks.
        for callback in callbacks:
            # Figure out which args need loading.
            items = callback[dd.Input] + callback[dd.State]
            item_ids = [_create_callback_id(item) for item in items]
            serverside_outputs = [serverside_output_map.get(item_id, None) for item_id in item_ids]
            # If any arguments are packed, unpack them.
            if any(serverside_outputs):
                f = callback["f"]
                callback["f"] = _unpack_outputs(serverside_outputs)(f)
        # 3) Apply the caching itself.
        for i, callback in enumerate(serverside_callbacks):
            f = callback["f"]
            callback["f"] = _pack_outputs(callback)(f)
        # 4) Strip special args.
        for callback in callbacks:
            for key in ["memoize"]:
                callback["kwargs"].pop(key, None)

        return callbacks


def _unpack_outputs(serverside_outputs):
    def unpack(f):
        @functools.wraps(f)
        def decorated_function(*args):
            if not any(serverside_outputs):
                return f(*args)
            args = list(args)
            for i, serverside_output in enumerate(serverside_outputs):
                # Just skip elements that are not stored server side.
                if not serverside_output:
                    continue
                # Replace content of element(s).
                try:
                    args[i] = serverside_output.backend.get(args[i], ignore_expired=True)
                except TypeError as ex:
                    # TODO: Should we do anything about this?
                    args[i] = None
            return f(*args)

        return decorated_function

    return unpack


def _pack_outputs(callback):
    memoize = callback["kwargs"].get("memoize", None)

    def packed_callback(f):
        @functools.wraps(f)
        def decorated_function(*args):
            multi_output = callback["multi_output"]
            # If memoize is enabled, we check if the cache already has a valid value.
            if memoize:
                # Figure out if an update is necessary.
                unique_ids = []
                update_needed = False
                for i, output in enumerate(callback[Output]):
                    # Filter out Triggers (a little ugly to do here, should ideally be handled elsewhere).
                    is_trigger = trigger_filter(callback["sorted_args"])
                    filtered_args = [arg for i, arg in enumerate(args) if not is_trigger[i]]
                    # Generate unique ID.
                    unique_id = _get_cache_id(f, output, list(filtered_args), output.session_check, output.arg_check)
                    unique_ids.append(unique_id)
                    if not output.backend.has(unique_id):
                        update_needed = True
                        break
                # If not update is needed, just return the ids (or values, if not serverside output).
                if not update_needed:
                    results = [uid if isinstance(callback[Output][i], ServersideOutput) else
                               callback[Output][i].backend.get(uid) for i, uid in enumerate(unique_ids)]
                    return results if multi_output else results[0]
            # Do the update.
            data = f(*args)
            data = list(data) if multi_output else [data]
            if callable(memoize):
                data = memoize(data)
            for i, output in enumerate(callback[Output]):
                # Skip no_update updates.
                if isinstance(data[i], type(dash.no_update)):
                    continue
                # Replace only for server side outputs.
                serverside_output = isinstance(callback[Output][i], ServersideOutput)
                if serverside_output or memoize:
                    # Filter out Triggers (a little ugly to do here, should ideally be handled elsewhere).
                    is_trigger = trigger_filter(callback["sorted_args"])
                    filtered_args = [arg for i, arg in enumerate(args) if not is_trigger[i]]
                    unique_id = _get_cache_id(f, output, list(filtered_args), output.session_check, output.arg_check)
                    output.backend.set(unique_id, data[i])
                    # Replace only for server side outputs.
                    if serverside_output:
                        data[i] = unique_id
            return data if multi_output else data[0]

        return decorated_function

    return packed_callback


def _get_cache_id(func, output, args, session_check=None, arg_check=True):
    all_args = [func.__name__, _create_callback_id(output)]
    if arg_check:
        all_args += list(args)
    if session_check:
        all_args += [_get_session_id()]
    return hashlib.md5(json.dumps(all_args).encode()).hexdigest()


def _get_output_id(callback):
    if isinstance(callback['f'], (ClientsideFunction, str)):
        f_repr = repr(callback['f'])  # handles clientside functions
    else:
        f_repr = f"{callback['f'].__module__}.{callback['f'].__name__}"  # handles Python functions
    f_hash = hashlib.md5(f_repr.encode()).digest()
    return str(uuid.UUID(bytes=f_hash, version=4))


# Interface definition for server stores.

class ServerStore:

    def get(self, key, ignore_expired=False):
        raise NotImplementedError()

    def set(self, key, value):
        raise NotImplementedError()

    def has(self, key):
        raise NotImplementedError()


# Place store implementations here.

class FileSystemStore(FileSystemCache):

    def __init__(self, cache_dir="file_system_store", **kwargs):
        super().__init__(cache_dir, **kwargs)

    def get(self, key, ignore_expired=False):
        if not ignore_expired:
            return super().get(key)
        # TODO: This part must be implemented for each type of cache.
        filename = self._get_filename(key)
        try:
            with open(filename, "rb") as f:
                pickle_time = pickle.load(f)  # ignore time
                return pickle.load(f)
        except (IOError, OSError, pickle.PickleError):
            return None


class RedisStore(RedisCache):

    def __init__(self, default_timeout=24 * 3600, **kwargs):
        """
        The timeout must be large enough that a (k,v) pair NEVER expires during a user session.
        """
        super().__init__(default_timeout=default_timeout, **kwargs)

    def get(self, key, ignore_expired=False):
        # TODO: Is there any way to honor ignore_expired for redis? I don't think so
        return super().get(key)


# endregion

# region No output transform

class NoOutputTransform(DashTransform):

    def __init__(self):
        self.initialized = False
        self.hidden_divs = []

    def layout(self, layout, layout_is_function):
        if layout_is_function or not self.initialized:
            children = _as_list(layout.children) + self.hidden_divs
            layout.children = children
            self.initialized = True
        return layout

    def _apply(self, callbacks):
        for callback in callbacks:
            if len(callback[dd.Output]) == 0:
                output_id = _get_output_id(callback)
                hidden_div = html.Div(id=output_id, style={"display": "none"})
                callback[dd.Output] = [dd.Output(output_id, "children")]
                self.hidden_divs.append(hidden_div)
        return callbacks

    def apply_serverside(self, callbacks):
        return self._apply(callbacks)

    def apply_clientside(self, callbacks):
        return self._apply(callbacks)


# endregion

# region Transformer implementations

class Dash(DashProxy):
    def __init__(self, *args, output_defaults=None, **kwargs):
        output_defaults = dict(backend=None, session_check=True) if output_defaults is None else output_defaults
        transforms = [TriggerTransform(), MultiplexerTransform(), NoOutputTransform(),
                      ServersideOutputTransform(**output_defaults)]
        super().__init__(*args, transforms=transforms, **kwargs)

# endregion



import io
import json
import ntpath
import base64
import uuid
import dash
import dash_html_components as html

from dash.dependencies import Output, Input


# region Utils for Download component


def send_file(path, filename=None, mime_type=None):
    """
    Convert a file into the format expected by the Download component.
    :param path: path to the file to be sent
    :param filename: name of the file, if not provided the original filename is used
    :param mime_type: mime type of the file (optional, passed to Blob in the javascript layer)
    :return: dict of file content (base64 encoded) and meta data used by the Download component
    """
    # If filename is not set, read it from the path.
    if filename is None:
        filename = ntpath.basename(path)
    # Read the file into a base64 string.
    with open(path, 'rb') as f:
        content = base64.b64encode(f.read()).decode()
    # Wrap in dict.
    return dict(content=content, filename=filename, mime_type=mime_type, base64=True)


def send_bytes(writer, filename, mime_type=None, **kwargs):
    """
    Convert data written to BytesIO into the format expected by the Download component.
    :param writer: a writer that can write to BytesIO
    :param filename: the name of the file
    :param mime_type: mime type of the file (optional, passed to Blob in the javascript layer)
    :return: dict of data frame content (base64 encoded) and meta data used by the Download component
    """
    data_io = io.BytesIO()
    # Some pandas writers try to close the IO, we do not want that.
    data_io_close = data_io.close
    data_io.close = lambda: None
    # Write data content to base64 string.
    writer(data_io, **kwargs)
    data_value = data_io.getvalue()
    data_io_close()
    content = base64.b64encode(data_value).decode()
    # Wrap in dict.
    return dict(content=content, filename=filename, mime_type=mime_type, base64=True)


def send_string(writer, filename, mime_type=None, **kwargs):
    """
    Convert data written to StringIO into the format expected by the Download component.
    :param writer: a writer that can write to StringIO
    :param filename: the name of the file
    :param mime_type: mime type of the file (optional, passed to Blob in the javascript layer)
    :return: dict of data frame content (base64 encoded) and meta data used by the Download component
    """
    data_io = io.StringIO()
    # Some pandas writers try to close the IO, we do not want that.
    data_io_close = data_io.close
    data_io.close = lambda: None
    # Write data content to base64 string.
    writer(data_io, **kwargs)
    data_value = data_io.getvalue().encode()
    data_io_close()
    content = base64.b64encode(data_value).decode()
    # Wrap in dict.
    return dict(content=content, filename=filename, mime_type=mime_type, base64=True)


known_pandas_writers = {
    "to_csv": False,
    "to_json": False,
    "to_html": False,
    "to_excel": True,
    "to_hdf": True,
    "to_feather": True,
    "to_parquet": True,
    "to_msgpack": True,
    "to_stata": True,
    "to_pickle": True,
}


def send_data_frame(writer, filename, mime_type=None, **kwargs):
    """
    Convert data frame into the format expected by the Download component.
    :param writer: a data frame writer
    :param filename: the name of the file
    :param mime_type: mime type of the file (optional, passed to Blob in the javascript layer)
    :return: dict of data frame content (base64 encoded) and meta data used by the Download component

    Examples
    --------

    >>> df = pd.DataFrame({'a': [1, 2, 3, 4], 'b': [2, 1, 5, 6], 'c': ['x', 'x', 'y', 'y']})
    ...
    >>> send_data_frame(df.to_csv, "mydf.csv")  # download as csv
    >>> send_data_frame(df.to_json, "mydf.json")  # download as json
    >>> send_data_frame(df.to_excel, "mydf.xls", index=False) # download as excel
    >>> send_data_frame(df.to_pkl, "mydf.pkl") # download as pickle

    """
    name = writer.__name__
    # Check if the provided writer is known.
    if name not in known_pandas_writers.keys():
        raise ValueError("The provided writer ({}) is not supported, "
                         "try calling send_string or send_bytes directly.".format(name))
    # If binary, use send_bytes.
    if known_pandas_writers[name]:
        return send_bytes(writer, filename, mime_type, **kwargs)
    # Otherwise, use send_string.
    return send_string(writer, filename, mime_type, **kwargs)


# endregion

# region Get triggered


class Triggered(object):
    def __init__(self, id, **kwargs):
        self.id = id
        for key in kwargs:
            setattr(self, key, kwargs[key])


def get_triggered() -> Triggered:
    triggered = dash.callback_context.triggered
    if not triggered:
        return Triggered(None)
    # Collect trigger ids and values.
    triggered_id = None
    triggered_values = {}
    for entry in triggered:
        # TODO: Test this part.
        elements = entry['prop_id'].split(".")
        current_id = ".".join(elements[:-1])
        current_prop = elements[-1]
        # Determine the trigger object.
        if triggered_id is None:
            triggered_id = current_id
        # TODO: Should all properties of the trigger be registered, or only one?
        if triggered_id != current_id:
            continue
        triggered_values[current_prop] = entry['value']
    # Now, create an object.
    try:
        triggered_id = json.loads(triggered_id)
    except ValueError:
        pass
    return Triggered(triggered_id, **triggered_values)


# endregion

def fix_page_load_anchor_issue(app, delay=None):
    """
    Fixes the issue that the pages is not scrolled to the anchor position on initial load.
    :param app: the Dash app object
    :param delay: in some cases, an additional delay might be needed for the page to load, specify in ms
    :return: dummy elements, which must be added to the layout for the fix to work
    """
    # Create dummy components.
    input_id, output_id = str(uuid.uuid4()), str(uuid.uuid4())
    dummy_input = html.Div(id=input_id, style={"display": "hidden"})
    dummy_output = html.Div(id=output_id, style={"display": "hidden"})
    # Setup the callback that does the magic.
    app.clientside_callback(
        """
        function(dummy_value) {{
            setTimeout(function(){{
                const match = document.getElementById(window.location.hash.substring(1))
                match.scrollIntoView();
            }}, {});
        }}
        """.format(delay),
        Output(output_id, "children"), [Input(input_id, "children")], prevent_initial_call=False)
    return [dummy_input, dummy_output]



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Keyboard(Component):
    """A Keyboard component.
The Keyboard component listens for keyboard events.

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- captureKeys (list of strings; optional):
    The keys to capture. Defaults to all keys.

- eventProps (list of strings; default ["key", "altKey", "ctrlKey", "shiftKey","metaKey", "repeat"]):
    The event properties to forward to dash, see
    https://developer.mozilla.org/en-US/docs/Web/API/KeyboardEvent.

- keydown (dict; optional):
    keydown (dict) the object that holds the result of the key down
    event. It is a dictionary with the following keys:      \"key\",
    \"altKey\", \"ctrlKey\", \"shiftKey\",\"metaKey\", \"repeat\".
    Those keys have the following values:     - key (str) which key is
    pressed    - altKey (bool) whether the Alt key is pressed    -
    ctrlKey (bool) Ctrl key is pressed    - shiftKey (bool) Shift key
    is pressed    - metaKey (bool) Meta key is pressed (Mac: Command
    key or PC: Windows key)    - repeat (bool) whether the key is held
    down.

- keys_pressed (dict; optional):
    keys_pressed (dict) is a dict of objects like keydown for all keys
    currently pressed.

- keyup (dict; optional):
    keyup (dict) the object that holds the result of the key up event.
    Structure like keydown.

- n_keydowns (number; default 0):
    A counter, which is incremented on each key down event, similar to
    n_clicks for buttons.

- n_keyups (number; default 0):
    A counter, which is incremented on each key up event, similar to
    n_clicks for buttons."""
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, eventProps=Component.UNDEFINED, captureKeys=Component.UNDEFINED, keydown=Component.UNDEFINED, keyup=Component.UNDEFINED, keys_pressed=Component.UNDEFINED, n_keydowns=Component.UNDEFINED, n_keyups=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'captureKeys', 'eventProps', 'keydown', 'keys_pressed', 'keyup', 'n_keydowns', 'n_keyups']
        self._type = 'Keyboard'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'captureKeys', 'eventProps', 'keydown', 'keys_pressed', 'keyup', 'n_keydowns', 'n_keyups']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Keyboard, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class MyTicker(Component):
    """A MyTicker component.
A modified version of dcc.Link that adds a few more options. E.g. you can disable scrolling to
the top upon updating the url.

Keyword arguments:
- children (a list of or a singular dash component, string or number; optional): The children of this component
- direction (a value equal to: "toRight", "toLeft"; optional)
- mode (a value equal to: "chain", "await", "smooth"; optional)
- move (boolean; optional)
- offset (a value equal to: PropTypes.number, PropTypes.string; optional)
- speed (number; optional)
- height (a value equal to: PropTypes.number, PropTypes.string; optional)
- id (string; optional): The ID used to identify this component in Dash callbacks
- className (string; optional): The class of the component"""
    @_explicitize_args
    def __init__(self, children=None, direction=Component.UNDEFINED, mode=Component.UNDEFINED, move=Component.UNDEFINED, offset=Component.UNDEFINED, speed=Component.UNDEFINED, height=Component.UNDEFINED, id=Component.UNDEFINED, className=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'direction', 'mode', 'move', 'offset', 'speed', 'height', 'id', 'className']
        self._type = 'MyTicker'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'direction', 'mode', 'move', 'offset', 'speed', 'height', 'id', 'className']
        self.available_wildcard_properties =            []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(MyTicker, self).__init__(children=children, **args)



import os
import re
import json
import dash

from distutils.dir_util import copy_tree
from flask import request, jsonify, make_response

'''
This purpose of this module is to ease the integration of Dash with dataiku. 
'''


def parse_config(args):
    return {'webAppBackendUrl': args.get('webAppBackendUrl'), 'appPrefix': args.get('appPrefix')}


def get_dataiku_kwargs(server, config_path):
    try:
        with open(config_path, 'r') as f:
            config = json.load(f)
        app_prefix = config['appPrefix']
        web_app_backend_url = config['webAppBackendUrl']
        return {
            'server': server,
            'routes_pathname_prefix': f'/{app_prefix}/',
            'requests_pathname_prefix': f"{web_app_backend_url}{app_prefix}/"
        }
    except FileNotFoundError:
        return {'server': server}


def setup_dataiku(server, config_path):
    # Add config route.
    @server.route("/configure")
    def configure():
        config = parse_config(request.args)
        # Check if the configuration has changed.
        if os.path.isfile(config_path):
            with open(config_path, 'r') as f:
                current_config = json.load(f)
            # Configuration has not changed, redirect to app.
            if config == current_config:
                return jsonify(success=True)
        # Configuration changed. Write new config and ask for restart.
        with open(config_path, 'w') as f:
            json.dump(config, f)
        return make_response(jsonify({'error': 'Configuration changed. Backend restart required.'}), 500)

    # Return keyword arguments.
    try:
        with open(config_path, 'r') as f:
            config = json.load(f)
        app_prefix = config['appPrefix']
        web_app_backend_url = config['webAppBackendUrl']
        return {
            'server': server,
            'routes_pathname_prefix': f'/{app_prefix}/',
            'requests_pathname_prefix': f"{web_app_backend_url}{app_prefix}/"
        }
    except FileNotFoundError:
        return {'server': server}


def bind_assets_folder(app: dash.Dash, app_id: str, assets_folder: str):
    # Create assets container folder for the app.
    dst_assets_folder = os.path.join(app.config.assets_folder, app_id)
    os.makedirs(dst_assets_folder, exist_ok=True)
    # Copy assets.
    copy_tree(assets_folder, dst_assets_folder)

    def _walk_assets_directory(self):
        walk_dir = os.path.join(self.config.assets_folder, app_id)  # EMHER: Use application sub dir
        slash_splitter = re.compile(r"[\\/]+")
        ignore_str = self.config.assets_ignore
        ignore_filter = re.compile(ignore_str) if ignore_str else None

        for current, _, files in sorted(os.walk(walk_dir)):
            if current == walk_dir:
                base = ""
            else:
                s = current.replace(walk_dir, "").lstrip("\\").lstrip("/")
                splitted = slash_splitter.split(s)
                if len(splitted) > 1:
                    base = "/".join(slash_splitter.split(s))
                else:
                    base = splitted[0]
            base = os.path.join(base, app_id)  # EMHER: Use application sub dir
            if ignore_filter:
                files_gen = (x for x in files if not ignore_filter.search(x))
            else:
                files_gen = files

            for f in sorted(files_gen):
                path = "/".join([base, f]) if base else f

                full = os.path.join(current, f)

                if f.endswith("js"):
                    self.scripts.append_script(self._add_assets_resource(path, full))
                elif f.endswith("css"):
                    self.css.append_css(self._add_assets_resource(path, full))
                elif f == "favicon.ico":
                    self._favicon = path

    app._walk_assets_directory = lambda x=app: _walk_assets_directory(app)



import dash_html_components as html
import dash_core_components as dcc

from dash.exceptions import PreventUpdate
from dash_extensions.enrich import Output, Input, PrefixIdTransform, DashProxy, prefix_recursively

URL_ID = "url"
CONTENT_ID = "content"


class Page:
    def __init__(self, id, label, layout=None, callbacks=None, prefix_ids=True, proxy=None):
        self.id = id
        self.label = label
        self._prefix_ids = prefix_ids
        self._layout = layout
        # Per default, use prefix transform.
        self._proxy = proxy
        if proxy is None:
            transforms = [PrefixIdTransform(id)] if prefix_ids else []
            self._proxy = DashProxy(transforms=transforms)
            if callbacks is not None:
                callbacks(self._proxy)

    @property
    def layout(self):
        def _layout(*args, **kwargs):
            if self._layout:
                self._proxy.layout = self._layout(*args, **kwargs)
                for transform in self._proxy.transforms:
                    if isinstance(transform, PrefixIdTransform):
                        transform.initialized = False
            return self._proxy._layout_value()  # layout

        return _layout

    @layout.setter
    def layout(self, layout):
        self._layout = layout

    @property
    def callbacks(self):
        def _callbacks(app):
            return self._proxy._register_callbacks(app)

        return _callbacks

    @callbacks.setter
    def callbacks(self, callbacks):
        callbacks(self._proxy)

    def render(self, layout):
        if self._prefix_ids:
            prefix_recursively(layout, self.id)
        return layout


class PageCollection:
    def __init__(self, pages, default_page_id=None, path_to_page=None, is_authorized=None, unauthorized_layout=None):
        self._pages = pages
        self.is_authorized = is_authorized
        self.unauthorized_layout = unauthorized_layout \
            if unauthorized_layout is not None else lambda x, *args, **kwargs: html.Div("Unauthorized.")
        self.path_to_page = path_to_page
        self.default_page_id = default_page_id if default_page_id is not None else pages[0].id

    def navigate_to(self, path, *args, **kwargs):
        if path is None:
            raise PreventUpdate
        # Locate the page. Maybe make this is more advanced later, for now just snap the id.
        page_ids = [page.id for page in self.pages]
        next_page_id = self.default_page_id
        if path is not None:
            page_id = path[1:] if self.path_to_page is None else self.path_to_page(path)
            if page_id in page_ids:
                next_page_id = page_id
        # Check if user is authorized.
        if self.is_authorized is not None:
            if not self.is_authorized(next_page_id):
                return self.unauthorized_layout(path, *args, **kwargs)
        # Return page layout.
        return self.pages[page_ids.index(next_page_id)].layout(path, *args, **kwargs)

    def navigation(self, app, content_id=CONTENT_ID, url_id=URL_ID):
        @app.callback(Output(content_id, "children"), [Input(url_id, "pathname")])
        def navigate_to(path):
            return self.navigate_to(path)

    @property
    def pages(self):
        if self.is_authorized is None:
            return self._pages
        return [page for page in self._pages if self.is_authorized(page.id)]

    def callbacks(self, app):
        for page in self.pages:
            page.callbacks(app)


# region Conversions

def app_to_page(app, id, label):
    app.transforms.append(PrefixIdTransform(id))
    return Page(id=id, label=label, proxy=app)


def module_to_page(module, id, label, **kwargs):
    return Page(id=id, label=label, layout=module.layout, callbacks=module.callbacks, **kwargs)


# endregion

# region Example layouts

def default_layout(*args):
    return html.Div([html.Div(id=CONTENT_ID), dcc.Location(id=URL_ID)] + list(args))

# endregion



import uuid
import json
import secrets
from flask_sockets import Sockets
from flask import session


def ensure_session_id():
    if "socket_id" not in session:
        session["socket_id"] = uuid.uuid4()


def add_ws_endpoint(sockets, pool, handler, endpoint):
    @sockets.route(endpoint)
    def open_socket(ws):
        pool[session["socket_id"]] = ws
        while not ws.closed:
            handler(ws)


class SocketPool:

    def __init__(self, app, handler=lambda ws: ws.receive(), endpoint="/ws"):
        # Set session secret.
        if not app.server.secret_key:
            app.server.secret_key = secrets.token_urlsafe(16)
        app.server.before_request(ensure_session_id)
        # Create sockets.
        self.sockets = Sockets(app.server)
        self.pool = {}
        # Add endpoint.
        add_ws_endpoint(self.sockets, self.pool, handler, endpoint)

    def send(self, message):
        self._try_send(message, session["socket_id"])

    def broadcast(self, message):
        for client_id in list(self.pool.keys()):
            self._try_send(message, client_id)

    def _try_send(self, message, client_id):
        ws = self.pool[client_id]
        if ws.closed:
            return self.pool.pop(client_id)
        ws.send(json.dumps(message))


def run_server(app, port=5000):
    from gevent import pywsgi
    from geventwebsocket.handler import WebSocketHandler
    server = pywsgi.WSGIServer(('', port), app.server, handler_class=WebSocketHandler)
    server.serve_forever()



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class DeferScript(Component):
    """A DeferScript component.
Used to delay import of js resources until after React had been loaded. Typically used to apply js to dynamic
content. Based on https://github.com/Grasia/grasia-dash-components/blob/master/src/components/Import.react.js

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- src (string; optional):
    Local or external source of the javascript to import."""
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, src=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'src']
        self._type = 'DeferScript'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'src']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(DeferScript, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class BurgerMenu(Component):
    """A BurgerMenu component.


Keyword arguments:

- children (a list of or a singular dash component, string or number; optional):
    The children of this component.

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- className (string; optional):
    The class of the component.

- customBurgerIcon (boolean; optional)

- customCrossIcon (boolean; optional)

- disableAutoFocus (boolean; optional)

- disableCloseOnEsc (boolean; optional)

- disableOverlayClick (boolean; optional)

- effect (a value equal to: "slide", "stack", "elastic", "bubble", "push", "pushRotate", "scaleDown", "scaleRotate", "fallDown", "reveal"; default "slide")

- height (string; default "100%")

- isOpen (boolean; optional)

- noOverlay (boolean; optional)

- noTransition (boolean; optional)

- outerContainerId (string; optional)

- pageWrapId (string; optional)

- right (boolean; optional)

- style (dict; optional)

- width (string; default "300px")"""
    @_explicitize_args
    def __init__(self, children=None, width=Component.UNDEFINED, height=Component.UNDEFINED, effect=Component.UNDEFINED, pageWrapId=Component.UNDEFINED, outerContainerId=Component.UNDEFINED, right=Component.UNDEFINED, disableCloseOnEsc=Component.UNDEFINED, noOverlay=Component.UNDEFINED, disableOverlayClick=Component.UNDEFINED, noTransition=Component.UNDEFINED, customBurgerIcon=Component.UNDEFINED, customCrossIcon=Component.UNDEFINED, disableAutoFocus=Component.UNDEFINED, style=Component.UNDEFINED, isOpen=Component.UNDEFINED, id=Component.UNDEFINED, className=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'id', 'className', 'customBurgerIcon', 'customCrossIcon', 'disableAutoFocus', 'disableCloseOnEsc', 'disableOverlayClick', 'effect', 'height', 'isOpen', 'noOverlay', 'noTransition', 'outerContainerId', 'pageWrapId', 'right', 'style', 'width']
        self._type = 'BurgerMenu'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'id', 'className', 'customBurgerIcon', 'customCrossIcon', 'disableAutoFocus', 'disableCloseOnEsc', 'disableOverlayClick', 'effect', 'height', 'isOpen', 'noOverlay', 'noTransition', 'outerContainerId', 'pageWrapId', 'right', 'style', 'width']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(BurgerMenu, self).__init__(children=children, **args)



from .BeforeAfter import BeforeAfter
from .Burger import Burger
from .DeferScript import DeferScript
from .Download import Download
from .EventListener import EventListener
from .Keyboard import Keyboard
from .Lottie import Lottie
from .Mermaid import Mermaid
from .Monitor import Monitor
from .Purify import Purify
from .Ticker import Ticker
from .WebSocket import WebSocket

__all__ = [
    "BeforeAfter",
    "Burger",
    "DeferScript",
    "Download",
    "EventListener",
    "Keyboard",
    "Lottie",
    "Mermaid",
    "Monitor",
    "Purify",
    "Ticker",
    "WebSocket"
]


# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Mermaid(Component):
    """A Mermaid component.


Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- chart (string; optional):
    The mermaid code of your chart. Check Mermaid js documentation for
    details.

- className (string; optional):
    The class of the component.

- config (dict; optional):
    On optional object with one of several Mermaid config parameters.
    Check Mermaid js documentation for details.

- name (string; default makeId(5)):
    On optional name of your mermaid diagram/flowchart/gantt etc."""
    @_explicitize_args
    def __init__(self, chart=Component.UNDEFINED, name=Component.UNDEFINED, config=Component.UNDEFINED, id=Component.UNDEFINED, className=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'chart', 'className', 'config', 'name']
        self._type = 'Mermaid'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'chart', 'className', 'config', 'name']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Mermaid, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Download(Component):
    """A Download component.
The Download component opens a download dialog when the data property (dict of filename, content, and type) changes.

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- base64 (boolean; default False):
    Default value for base64.

- data (dict; optional):
    When set, a download is invoked using a Blob.

    `data` is a dict with keys:

    - base64 (boolean; optional)

    - content (boolean | number | string | dict | list; required)

    - filename (string; required)

    - mime_type (string; optional)

- mime_type (string; default "text/plain"):
    Default value for mime_type."""
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, data=Component.UNDEFINED, base64=Component.UNDEFINED, mime_type=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'base64', 'data', 'mime_type']
        self._type = 'Download'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'base64', 'data', 'mime_type']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Download, self).__init__(**args)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Monitor(Component):
    """A Monitor component.
The Sync component makes it possible to synchronize states between components.

Keyword arguments:

- children (list of a list of or a singular dash component, string or numbers; optional):
    The children of this component. Must be a list of components with
    length > 1.

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- className (string; optional):
    A custom class name.

- data (dict; optional)

- probes (dict; optional):
    List of probes. Each link is a list of dicts that specify which
    properties each probe records.

    `probes` is a dict with strings as keys and values of type list of
    list of strings | dict with keys:

    - id (string | dict; optional)

    - prop (boolean | number | string | dict | list; required)s

- style (dict; optional):
    The CSS style of the component."""
    @_explicitize_args
    def __init__(self, children=None, probes=Component.UNDEFINED, data=Component.UNDEFINED, id=Component.UNDEFINED, style=Component.UNDEFINED, className=Component.UNDEFINED, **kwargs):
        self._prop_names = ['children', 'id', 'className', 'data', 'probes', 'style']
        self._type = 'Monitor'
        self._namespace = 'dash_extensions'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['children', 'id', 'className', 'data', 'probes', 'style']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}
        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Monitor, self).__init__(children=children, **args)
